"""
The code in this module is adapted from https://github.com/pantsbuild/pex/blob/master/pex/pex.py

It is used to enter an interactive interpreter session from an executable created with ``shiv``.
"""
import code
import sys


def _exec_function(ast, globals_map):
    locals_map = globals_map
    exec(ast, globals_map, locals_map)
    return locals_map


def execute_content(name, content):
    try:
        ast = compile(content, name, "exec", flags=0, dont_inherit=1)
    except SyntaxError:
        raise RuntimeError(f"Unable to parse {name}. Is it a Python script? Syntax correct?")

    old_name, old_file = globals().get("__name__"), globals().get("__file__")

    try:
        globals()["__name__"] = "__main__"
        globals()["__file__"] = name
        _exec_function(ast, globals())
    finally:
        if old_name:
            globals()["__name__"] = old_name
        else:
            globals().pop("__name__")
        if old_file:
            globals()["__file__"] = old_file
        else:
            globals().pop("__file__")


def execute_interpreter():
    if sys.argv[1:]:
        try:
            with open(sys.argv[1]) as fp:
                name, content = sys.argv[1], fp.read()
        except (FileNotFoundError, IsADirectoryError, PermissionError) as e:
            raise RuntimeError(
                f"Could not open {sys.argv[1]} in the environment [{sys.argv[0]}]: {e}"
            )

        sys.argv = sys.argv[1:]
        execute_content(name, content)
    else:
        code.interact()



import os

from shiv.constants import PIP_REQUIRE_VIRTUALENV
from shiv.pip import clean_pip_env


def test_clean_pip_env(monkeypatch):

    before_env_var = 'test'
    monkeypatch.setenv(PIP_REQUIRE_VIRTUALENV, before_env_var)

    with clean_pip_env():
        assert PIP_REQUIRE_VIRTUALENV not in os.environ

    assert os.environ.get(PIP_REQUIRE_VIRTUALENV) == before_env_var



from setuptools import setup
import setuptools


setup(
    name='hello',
    packages=['hello'],
    package_data={'': ['script.sh']},
    entry_points={"console_scripts": ["hello = hello:main"]},
)



import contextlib
import json
import os
import subprocess
import sys
import stat
import tempfile

from pathlib import Path

import pytest

from click.testing import CliRunner

from shiv.cli import main, find_entry_point, _interpreter_path
from shiv.constants import DISALLOWED_PIP_ARGS, NO_PIP_ARGS_OR_SITE_PACKAGES, NO_OUTFILE, DISALLOWED_ARGS
from shiv.info import main as info_main
from shiv.pip import install

UGOX = stat.S_IXUSR | stat.S_IXGRP | stat.S_IXOTH


@contextlib.contextmanager
def mocked_sys_prefix():
    attribute_to_mock = "real_prefix" if hasattr(sys, "real_prefix") else "base_prefix"
    original = getattr(sys, attribute_to_mock)
    setattr(sys, attribute_to_mock, "/fake/dir")
    yield
    setattr(sys, attribute_to_mock, original)


class TestCLI:

    @pytest.fixture
    def shiv_root(self, monkeypatch, tmpdir):

        with tempfile.TemporaryDirectory(dir=tmpdir) as tmpdir:
            os.environ['SHIV_ROOT'] = tmpdir
            yield tmpdir

        os.environ.pop("SHIV_ROOT")

    @pytest.fixture
    def runner(self):
        """Returns a click test runner."""

        return lambda args: CliRunner().invoke(main, args)

    @pytest.fixture
    def info_runner(self):
        """Returns a click test runner (for shiv-info)."""

        return lambda args: CliRunner().invoke(info_main, args)

    def test_find_entry_point(self, tmpdir, package_location):
        """Test that we can find console_script metadata."""
        install(['-t', str(tmpdir), str(package_location)])
        assert find_entry_point(Path(tmpdir), "hello") == "hello:main"

    def test_no_args(self, runner):
        """This should fail with a warning about supplying pip arguments"""

        result = runner([])

        assert result.exit_code == 1
        assert NO_PIP_ARGS_OR_SITE_PACKAGES in result.output

    def test_no_outfile(self, runner):
        """This should fail with a warning about not providing an outfile"""

        result = runner(['-e', 'test', 'flask'])

        assert result.exit_code == 1
        assert NO_OUTFILE in result.output

    def test_find_interpreter(self):

        interpreter = _interpreter_path()

        assert Path(interpreter).exists()
        assert Path(interpreter).is_file()

    def test_find_interpreter_false(self):

        with mocked_sys_prefix():
            interpreter = _interpreter_path()

        # should fall back on the current sys.executable
        assert interpreter == sys.executable

    @pytest.mark.parametrize("arg", [arg for tup in DISALLOWED_ARGS.keys() for arg in tup])
    def test_disallowed_args(self, runner, arg):
        """This method tests that all the potential disallowed arguments match their error messages."""

        # run shiv with a disallowed argument
        result = runner(['-o', 'tmp', arg])

        # get the 'reason' message:
        for disallowed in DISALLOWED_ARGS:
            if arg in disallowed:
                reason = DISALLOWED_ARGS[disallowed]

        assert result.exit_code == 1

        # assert we got the correct reason
        assert DISALLOWED_PIP_ARGS.format(arg=arg, reason=reason) in result.output

    @pytest.mark.parametrize('compile_option', ["--compile-pyc", "--no-compile-pyc"])
    @pytest.mark.parametrize('force', ["yes", "no"])
    def test_hello_world(self, runner, info_runner, shiv_root, package_location, compile_option, force):
        output_file = Path(shiv_root, 'test.pyz')

        result = runner(['-e', 'hello:main', '-o', str(output_file), str(package_location), compile_option])

        # check that the command successfully completed
        assert result.exit_code == 0

        # ensure the created file actually exists
        assert output_file.exists()

        # build env
        env = {**os.environ, "SHIV_FORCE_EXTRACT": force}

        # now run the produced zipapp
        proc = subprocess.run([str(output_file)], stdout=subprocess.PIPE, shell=True, env=env)

        assert proc.stdout.decode() == "hello world" + os.linesep

        # now run shiv-info on the produced zipapp
        result = info_runner([str(output_file)])

        # check the rc and output
        assert result.exit_code == 0
        assert f"pyz file: {str(output_file)}" in result.output

        # ensure that executable permissions were retained (skip test on windows)
        if os.name != "nt":
            build_id = json.loads(info_runner([str(output_file), '--json']).output)['build_id']
            assert Path(shiv_root, f"{output_file.stem}_{build_id}", "site-packages", "hello", "script.sh").stat().st_mode & UGOX == UGOX

    @pytest.mark.parametrize("extend_path", ["--extend-pythonpath", "--no-extend-pythonpath", "-E"])
    def test_extend_pythonpath(self, shiv_root, runner, extend_path):

        output_file = Path(shiv_root, 'test_pythonpath.pyz')
        package_dir = Path(shiv_root, 'package')
        main_script = Path(package_dir, 'env.py')

        MAIN_PROG = '\n'.join([
            "import os",
            "def main():",
            "    print(os.environ.get('PYTHONPATH', ''))"
        ])

        package_dir.mkdir()
        main_script.write_text(MAIN_PROG)

        result = runner([
            '-e', 'env:main',
            '-o', str(output_file),
            '--site-packages', str(package_dir),
            extend_path,
        ])

        # check that the command successfully completed
        assert result.exit_code == 0

        # ensure the created file actually exists
        assert output_file.exists()

        # now run the produced zipapp and confirm shiv_root is in PYTHONPATH
        proc = subprocess.run([str(output_file)], stdout=subprocess.PIPE, shell=True, env=os.environ)

        pythonpath_has_root = str(shiv_root) in proc.stdout.decode()
        assert extend_path.startswith('--no') != pythonpath_has_root

    def test_no_entrypoint(self, shiv_root, runner, package_location, monkeypatch):

        output_file = Path(shiv_root, 'test.pyz')

        result = runner(['-o', str(output_file), str(package_location)])

        # check that the command successfully completed
        assert result.exit_code == 0

        # ensure the created file actually exists
        assert output_file.exists()

        # now run the produced zipapp
        proc = subprocess.run(
            [str(output_file)],
            input=b"import hello;print(hello)",
            stdout=subprocess.PIPE,
            stderr=subprocess.PIPE,
            shell=True,
            env=os.environ,
        )

        assert proc.returncode == 0
        assert "hello" in proc.stdout.decode()



import tempfile
import stat
import os
import sys
import zipfile

from pathlib import Path
from zipapp import ZipAppError

import pytest

from shiv.builder import write_file_prefix, create_archive


UGOX = stat.S_IXUSR | stat.S_IXGRP | stat.S_IXOTH


def tmp_write_prefix(interpreter):
    with tempfile.TemporaryFile() as fd:
        write_file_prefix(fd, interpreter)
        fd.seek(0)
        written = fd.read()

    return written


class TestBuilder:

    @pytest.mark.parametrize(
        "interpreter,expected",
        [
            ("/usr/bin/python", b"#!/usr/bin/python\n"),
            ("/usr/bin/env python", b"#!/usr/bin/env python\n"),
            ("/some/other/path/python -sE", b"#!/some/other/path/python -sE\n"),
        ],
    )
    def test_file_prefix(self, interpreter, expected):
        assert tmp_write_prefix(interpreter) == expected

    def test_binprm_error(self):
        with pytest.raises(SystemExit):
            tmp_write_prefix(f"/{'c' * 200}/python")

    def test_create_archive(self, sp):
        with tempfile.TemporaryDirectory() as tmpdir:
            target = Path(tmpdir, "test.zip")

            # create an archive
            create_archive(sp, target, sys.executable, "code:interact")

            # create one again (to ensure we overwrite)
            create_archive(sp, target, sys.executable, "code:interact")

            assert zipfile.is_zipfile(str(target))

            with pytest.raises(ZipAppError):
                create_archive(sp, target, sys.executable, "alsjdbas,,,")

    @pytest.mark.skipif(
        os.name == "nt", reason="windows has no concept of execute permissions"
    )
    def test_archive_permissions(self, sp):
        with tempfile.TemporaryDirectory() as tmpdir:
            target = Path(tmpdir, "test.zip")
            create_archive(sp, target, sys.executable, "code:interact")

            assert target.stat().st_mode & UGOX == UGOX



import click
import json
import zipfile


@click.command(context_settings=dict(help_option_names=["-h", "--help", "--halp"]))
@click.option("--json", "-j", "print_as_json", is_flag=True, help="output as plain json")
@click.argument("pyz")
def main(print_as_json, pyz):
    """A simple utility to print debugging information about PYZ files created with ``shiv``"""

    zip_file = zipfile.ZipFile(pyz)
    data = json.loads(zip_file.read("environment.json"))

    if print_as_json:
        click.echo(json.dumps(data, indent=4, sort_keys=True))

    else:
        click.echo()
        click.secho(f"pyz file: ", fg="green", bold=True, nl=False)
        click.secho(pyz, fg="white")
        click.echo()

        for key, value in data.items():
            click.secho(f"{key}: ", fg="blue", bold=True, nl=False)
            click.secho(f"{value}", fg="white")

        click.echo()



import os

from pathlib import Path

import pytest


@pytest.fixture
def zip_location():
    return Path(__file__).absolute().parent / 'test.zip'


@pytest.fixture(params=[True, False], ids=['.', 'absolute-path'])
def package_location(request):
    package_location = Path(__file__).absolute().parent / 'package'

    if request.param is True:
        # test building from the current directory
        cwd = os.getcwd()
        os.chdir(package_location)
        yield Path('.')
        os.chdir(cwd)
    else:
        # test building an absolute path
        yield package_location


@pytest.fixture
def sp():
    return Path(__file__).absolute().parent / 'sp' / 'site-packages'
